Attribute VB_Name = "modSessionTable"
Option Explicit
Option Base 1
      
Public SESSION_DISABLE_BOT_SESSIONS     As Boolean
Public SESSION_DISABLE_TOURIST_SESSIONS As Boolean

'***************************************************************************
'** Session Table store structure                                         **
'***************************************************************************
Public Type STRUCT_AW_SESSION
  Number                    As Long
  name                      As String
  Citnum                    As Long
  PrivNum                   As Long
  X                         As Long
  Y                         As Long
  Z                         As Long
  TerrainEnabled            As Boolean
  LastCommandTime           As Long
  DisableWarning            As Boolean
  IsBot                     As Boolean
  IsTourist                 As Boolean
End Type

'***************************************************************************
'** Public Enums                                                          **
'***************************************************************************
Public Enum SESSION_TIMER_TYPE
  SESSION_TIMER_COMMAND
End Enum

'***************************************************************************
'** Operational Constants                                                 **
'***************************************************************************
Public Const SESSION_NO_SESSION_INDEX       As Long = 0

'***************************************************************************
'** Timer Related Constants                                               **
'***************************************************************************
Private Const SESSION_TIMER_ANCHOR_DATE     As String = "01-01-2002"
Private Const SESSION_TIMER_UNIT            As String = "s"
Public SESSION_DISABLE_WARN_INTERVAL        As Long
Public SESSION_DISABLE_INTERVAL             As Long

'***************************************************************************
'** Session Table Stores                                                  **
'***************************************************************************
Public mBotSession                          As Long
Private mSessionCount                       As Long
Private mTerrainSessionCount                As Long
Private mSessions()                         As STRUCT_AW_SESSION

Public Sub SessionClearSessionTable()

    'This sub clears the session table and resets the number of terrain sessions
    'Necessary in the case of a world disconnect

    SessionTableTerm
    mTerrainSessionCount = 0
    FormSetTerrainSessionCount 0
    FormSetAvatarCount mSessionCount

End Sub
'***************************************************************************
'** Initialise the session table                                          **
'***************************************************************************
Public Sub SessionTableInit()

  mAWSDK.aw_event_set AW_EVENT_AVATAR_ADD
  mAWSDK.aw_event_set AW_EVENT_AVATAR_CHANGE
  mAWSDK.aw_event_set AW_EVENT_AVATAR_DELETE

  mSessionCount = 0
  Erase mSessions
  
End Sub


'***************************************************************************
'** Terminate the session table                                           **
'***************************************************************************
Public Sub SessionTableTerm()

  mSessionCount = 0
  Erase mSessions

End Sub


'***************************************************************************
'** Avatar Add Event                                                      **
'***************************************************************************
Public Sub SessionEventAvatarAdd()

  Dim sessionIndex          As Long
  
  '*************************************************************************
  '** Check for the existence of a session of the citizen number          **
  '**                                                                     **
  '** This is required to cater for sporadic disconnects during operation **
  '** so that the terrain editing status and timers are not discarded as  **
  '** the bot is subsequently notified of the avatars existence when it   **
  '** reconnects to the world server                                      **
  '*************************************************************************
  'Update December 1, 2002 - upon further thought, the session number of an
  'avatar still in the world will be the same, just like the citizen number
  'so there's no reason to keep unique citizen numbers
  'COMMENTED LINE:
  'sessionIndex = SessionTableGetCitizenIndex(mAWSDK.aw_int(AW_AVATAR_CITIZEN))
  
  sessionIndex = SessionTableGetSessionIndex(mAWSDK.aw_int(AW_AVATAR_SESSION))
  
  If (sessionIndex = SESSION_NO_SESSION_INDEX) Then
    
    '***********************************************************************
    '** Expand the session table for a new session                        **
    '***********************************************************************
    mSessionCount = mSessionCount + 1
    ReDim Preserve mSessions(mSessionCount)
    sessionIndex = mSessionCount
    
    '***********************************************************************
    '** Default the values otherwise retained on a disconnect/reconnect   **
    '***********************************************************************
    With mSessions(mSessionCount)
      .Citnum = mAWSDK.aw_int(AW_AVATAR_CITIZEN)
      .TerrainEnabled = False
      .DisableWarning = False
      .LastCommandTime = 0
      
   
    
    
      '***********************************************************************
      '** Tell the user the current status of the terrain manager           **
      '***********************************************************************
      'only say if the user has build; else don't tell user anything
      If HasRight(.PrivNum, AW_WORLD_BUILD_RIGHT) Then
        TerrainCacheStateMessage mAWSDK.aw_int(AW_AVATAR_SESSION)
      End If
  
    End With
  
  End If
  
  '*************************************************************************
  '** Set the non-retained session values                                 **
  '*************************************************************************
  With mSessions(sessionIndex)
    .Number = mAWSDK.aw_int(AW_AVATAR_SESSION)
    .name = mAWSDK.aw_string(AW_AVATAR_NAME)
    .PrivNum = mAWSDK.aw_int(AW_AVATAR_PRIVILEGE)
    .X = mAWSDK.aw_int(AW_AVATAR_X)
    .Y = mAWSDK.aw_int(AW_AVATAR_Y)
    .Z = mAWSDK.aw_int(AW_AVATAR_Z)
    .IsBot = SessionTableIsBotName(.name)
    .IsTourist = SessionTableIsTouristName(.name)
    
    '***********************************************************************
    '** Log the new avatar's arrival                                      **
    '***********************************************************************
    If .IsBot Then
      FormWriteToLog "BOT '" & .name & "' (Priv = " & .PrivNum & ") identified"
      
    ElseIf .IsTourist Then
      FormWriteToLog "TOURIST '" & .name & "' identified"
      
    Else
      FormWriteToLog .Citnum & " '" & .name & "' " _
                   & IIf(.Citnum = .PrivNum, vbNullString, "(Priv = " & .PrivNum & ") ") _
                   & " identified"
  
    End If
  End With
  
  '*************************************************************************
  '** Update the avatar count on the form                                 **
  '*************************************************************************
  FormSetAvatarCount mSessionCount
    
End Sub

Public Function FormGetSessionCount()

    FormGetSessionCount = mSessionCount

End Function


'***************************************************************************
'** Avatar Change Event                                                   **
'***************************************************************************
Public Sub SessionEventAvatarChange()

  Static sessionIndex       As Long
  Static prevCellX          As Long
  Static prevCellZ          As Long
  Static newCellX           As Long
  Static newCellZ           As Long
  
  '*************************************************************************
  '** Locate the session in the session table, ignoring if unknown        **
  '*************************************************************************
  sessionIndex = SessionTableGetSessionIndex(mAWSDK.aw_int(AW_AVATAR_SESSION))
  If Not (sessionIndex = SESSION_NO_SESSION_INDEX) Then
    With mSessions(sessionIndex)
    
      '*********************************************************************
      '** Store the avatar's previous cell location for comparison        **
      '*********************************************************************
      prevCellX = AWCellFromCoord(.X, AW_CELL_CALC_VERTEX)
      prevCellZ = AWCellFromCoord(.Z, AW_CELL_CALC_VERTEX)
      
      '*********************************************************************
      '** Store the new avatar location and calculate the current cell    **
      '*********************************************************************
      .X = mAWSDK.aw_int(AW_AVATAR_X)
      .Y = mAWSDK.aw_int(AW_AVATAR_Y)
      .Z = mAWSDK.aw_int(AW_AVATAR_Z)
      
      newCellX = AWCellFromCoord(.X, AW_CELL_CALC_VERTEX)
      newCellZ = AWCellFromCoord(.Z, AW_CELL_CALC_VERTEX)
      
      '*********************************************************************
      '** If terrain is enable and the avatar has changec cell, queue the **
      '** surrounding cells for potential querying                        **
      '*********************************************************************
      If .TerrainEnabled And Not ((newCellX = prevCellX) And (newCellZ = prevCellZ)) Then
        TerrainQueueAddArea newCellX, newCellZ
          
      End If
    End With
  End If
  
End Sub


'***************************************************************************
'** Avatar Delete Event                                                   **
'***************************************************************************
Public Sub SessionEventAvatarDelete()

  Dim sessionIndex          As Long
  Dim shiftIndex            As Long
  
  '*************************************************************************
  '** Locate the session in the session table ignoring unknown sessions.  **
  '** Sadly, this is required because either the SDK or the world server  **
  '** sometimes kicks out a spurious avatar delete event when entering an **
  '** empty world.                                                        **
  '*************************************************************************
  sessionIndex = SessionTableGetSessionIndex(mAWSDK.aw_int(AW_AVATAR_SESSION))
  If Not (sessionIndex = SESSION_NO_SESSION_INDEX) Then
    
    '***********************************************************************
    '** If the session had terrain enabled, update the form terrain count **
    '***********************************************************************
    If mSessions(sessionIndex).TerrainEnabled Then
      mTerrainSessionCount = mTerrainSessionCount - 1
      
    End If
    
    '***********************************************************************
    '** Shift the subsequent sessions "down" the table and resize         **
    '***********************************************************************
    For shiftIndex = sessionIndex To mSessionCount - 1
      mSessions(shiftIndex) = mSessions(shiftIndex + 1)
      
    Next shiftIndex

    mSessionCount = mSessionCount - 1
    If (mSessionCount <= 0) Then
      Erase mSessions
    
    Else
      ReDim Preserve mSessions(mSessionCount)
    
    End If

    '***********************************************************************
    '** Log the event and update the form                                 **
    '***********************************************************************
    FormWriteToLog mAWSDK.aw_string(AW_AVATAR_NAME) & " left"

    FormSetAvatarCount mSessionCount
    FormSetTerrainSessionCount mTerrainSessionCount
    
  End If

End Sub

Public Function SessionStateMessageAll()
  'This function parses the session table to determine who has build and consolemessages
  'all people who can build, but not anyone else
  
  Dim sessionCounter As Integer
  
  For sessionCounter = 1 To mSessionCount
    With mSessions(sessionCounter)
    
      If HasRight(.PrivNum, AW_WORLD_BUILD_RIGHT) Then
        TerrainCacheStateMessage .Number
      End If
      
    End With
  Next

End Function

'***************************************************************************
'** Retrieve a session number's index in the session table                **
'***************************************************************************
Private Function SessionTableGetSessionIndex(ByVal sessionNumber As Long) As Long

  Static sessionIndex       As Long
  Static sessionFound       As Boolean
  
  sessionFound = False
  
  '*************************************************************************
  '** Scan the session table for a matching session number                **
  '*************************************************************************
  For sessionIndex = 1 To mSessionCount
    If (mSessions(sessionIndex).Number = sessionNumber) Then
      sessionFound = True
      Exit For
      
    End If
  Next sessionIndex
  
  '*************************************************************************
  '** If the session was not found, return the default (no index)         **
  '*************************************************************************
  If Not sessionFound Then
    sessionIndex = SESSION_NO_SESSION_INDEX
    
  End If
  
  SessionTableGetSessionIndex = sessionIndex
  
End Function


'***************************************************************************
'** Retrieve a citizen number's index in the session table                **
'***************************************************************************
Private Function SessionTableGetCitizenIndex(ByVal citizenNumber As Long) As Long

  Dim sessionIndex          As Long
  Dim sessionFound          As Boolean
  
  sessionFound = False
  
  '*************************************************************************
  '** Scan the session table for a matching citizen number                **
  '*************************************************************************
  For sessionIndex = 1 To mSessionCount
    If (mSessions(sessionIndex).Citnum = citizenNumber) Then
      sessionFound = True
      Exit For
      
    End If
  Next sessionIndex
  
  '*************************************************************************
  '** If the session was not found, return the default (no index)         **
  '*************************************************************************
  If Not sessionFound Then
    sessionIndex = SESSION_NO_SESSION_INDEX
    
  End If
  
  SessionTableGetCitizenIndex = sessionIndex

End Function


'***************************************************************************
'** Retrieve a session using its session number                           **
'***************************************************************************
Public Function SessionTableGetSession(ByVal sessionNumber As Long) As STRUCT_AW_SESSION

  Dim sessionIndex          As Long
  Dim requestSession        As STRUCT_AW_SESSION
  
  '*************************************************************************
  '** Locate the session table entry                                      **
  '*************************************************************************
  sessionIndex = SessionTableGetSessionIndex(sessionNumber)
  
  '*************************************************************************
  '** Create a blank session if the session was not found                 **
  '*************************************************************************
  If (sessionIndex = SESSION_NO_SESSION_INDEX) Then
    With requestSession
      .Citnum = 0
      .PrivNum = 0
      .name = vbNullString
      .Number = 0
      .X = 0
      .Y = 0
      .Z = 0
    End With
    
  Else
    '***********************************************************************
    '** Retrieve the session from the session table                       **
    '***********************************************************************
    requestSession = mSessions(sessionIndex)
    
  End If
  
  SessionTableGetSession = requestSession

End Function


'***************************************************************************
'** Enable/Disable terrain features for a session number                  **
'***************************************************************************
Public Function SessionTableEnableTerrain(ByVal sessionNumber As Long, ByVal enableTerrain As Boolean) As Boolean

  Dim sessionIndex          As Long
  Dim citizenIndex          As Long
  Dim trackingChanged       As Boolean
  
  trackingChanged = False
  
  '*************************************************************************
  '** Locate the session in the session table                             **
  '*************************************************************************
  sessionIndex = SessionTableGetSessionIndex(sessionNumber)
  
  '*************************************************************************
  '** Ignore unknown sessions                                             **
  '*************************************************************************
  If (sessionIndex = SESSION_NO_SESSION_INDEX) Then
    FormWriteToLog "Unknown session detected attempting to enable terrain building"
    
  Else
    
    With mSessions(sessionIndex)
      
      '*********************************************************************
      '** Ignore bots ... potential flood issues!!                        **
      '*********************************************************************
      If (SESSION_DISABLE_BOT_SESSIONS And .IsBot) Then
        FormWriteToLog "Bot detected attempting to enable terrain building (priv = " & .PrivNum & ")"
        TerrainCacheConsoleMessage mSessions(sessionIndex).Number, "Bots are not permitted to enable terrain features in this world."
        
      '*********************************************************************
      '** Ignore Tourists - potential vandalism                           **
      '*********************************************************************
      ElseIf (SESSION_DISABLE_TOURIST_SESSIONS And .IsTourist) Then
        FormWriteToLog "Tourist detected attempting to enable terrain building (priv = " & .PrivNum & ")"
        TerrainCacheConsoleMessage mSessions(sessionIndex).Number, "Tourists are not permitted to enable terrain features in this world."
        
      '*********************************************************************
      '** If the terrain value for the session is changing                **
      '*********************************************************************
      
      ElseIf Not (.TerrainEnabled = enableTerrain) Then
        
        '*******************************************************************
        '** Set the terrain value and update the counters and form        **
        '*******************************************************************
        .TerrainEnabled = enableTerrain
        .DisableWarning = False
        trackingChanged = True
        
        mTerrainSessionCount = mTerrainSessionCount + IIf(enableTerrain, 1, -1)
        FormSetTerrainSessionCount mTerrainSessionCount
        
        FormWriteToLog .Citnum & " ''" & .name & "' has " & IIf(enableTerrain, "enabled", "disabled") & " terrain"
        
        '*******************************************************************
        '** If the terrain is turned ON, queue a query request for the    **
        '** cells surrounding the avatar's current position               **
        '*******************************************************************
        If enableTerrain Then
          TerrainQueueAddArea AWCellFromCoord(.X, AW_CELL_CALC_VERTEX), _
                              AWCellFromCoord(.Z, AW_CELL_CALC_VERTEX)
                
        End If
      End If
    End With
  End If
  
  SessionTableEnableTerrain = trackingChanged
  
End Function


'***************************************************************************
'** Set a supported session timer to the current time                     **
'***************************************************************************
Public Sub SessionTableSetTimer(ByVal sessionNumber As Long, ByVal timerType As SESSION_TIMER_TYPE)

  Static sessionIndex       As Long
  Static timerValue         As Long
  
  '*************************************************************************
  '** Locate the session in the session table                             **
  '*************************************************************************
  sessionIndex = SessionTableGetSessionIndex(sessionNumber)
  If Not (sessionIndex = SESSION_NO_SESSION_INDEX) Then
    
    '***********************************************************************
    '** Caluclate the value of the timer NOW                              **
    '***********************************************************************
    timerValue = DateDiff(SESSION_TIMER_UNIT, SESSION_TIMER_ANCHOR_DATE, Now())
    
    With mSessions(sessionIndex)
      '*********************************************************************
      '** For a command timer, update the timer and turn off the auto-    **
      '** disable "has been warned" indicator                             **
      '*********************************************************************
      If (timerType = SESSION_TIMER_COMMAND) Then
        .LastCommandTime = timerValue
        .DisableWarning = False
        
      End If
    End With
  End If

End Sub


'***************************************************************************
'** Auto-disable any inactive sessions                                    **
'***************************************************************************
Public Sub SessionTableAutoDisable()

  Static sessionIndex       As Long
  Static nowTimer           As Long
  
  '*************************************************************************
  '** Quick exit if there are no sessions enabled for terrain             **
  '*************************************************************************
  If (mTerrainSessionCount = 0) Then Exit Sub
  
  '*************************************************************************
  '** Calculate the timer value NOW                                       **
  '*************************************************************************
  nowTimer = DateDiff(SESSION_TIMER_UNIT, SESSION_TIMER_ANCHOR_DATE, Now())
  
  '*************************************************************************
  '** Scan the entire session table for expired terrain status            **
  '*************************************************************************
  For sessionIndex = 1 To mSessionCount
    With mSessions(sessionIndex)
      
      '*********************************************************************
      '** Ignore any sessions where terrain is not enabled                **
      '*********************************************************************
      If .TerrainEnabled Then
      
        '*******************************************************************
        '** Identify if the status is under threat for the session        **
        '*******************************************************************
        If ((nowTimer - .LastCommandTime) >= SESSION_DISABLE_WARN_INTERVAL) Then
          
          '*****************************************************************
          '** Check to see if the disable interval has been passed        **
          '*****************************************************************
          If ((nowTimer - .LastCommandTime) < SESSION_DISABLE_INTERVAL) Then
            
            '***************************************************************
            '** Disable hasn't been reached, but no warning issued        **
            '***************************************************************
            If Not .DisableWarning Then
              
              '*************************************************************
              '** Issue a warning to the user of an iminent auto disable  **
              '*************************************************************
              .DisableWarning = True
              TerrainCacheConsoleMessage .Number, TERRAIN_DISABLE_WARN_MESSAGE & " in " _
                                                & SESSION_DISABLE_INTERVAL - SESSION_DISABLE_WARN_INTERVAL _
                                                & SESSION_TIMER_UNIT
            End If
          Else
            
            '***************************************************************
            '** Terrain features have timed out, disable for this session **
            '***************************************************************
            .TerrainEnabled = False
            TerrainCacheConsoleMessage .Number, TERRAIN_DISABLE_MESSAGE
            
            '***************************************************************
            '** Decrement the counter and update the form                 **
            '***************************************************************
            mTerrainSessionCount = mTerrainSessionCount - 1
            FormSetTerrainSessionCount mTerrainSessionCount
            
            FormWriteToLog .Citnum & " ''" & .name & "' has timed out"
          
          End If
        End If
      End If
    End With
  Next sessionIndex

End Sub

Private Function SessionTableIsBotName(ByVal avatarName As String) As Boolean

  SessionTableIsBotName = ((Left(avatarName, 1) = "[") And (Right(avatarName, 1) = "]"))
  
End Function

Private Function SessionTableIsTouristName(ByVal avatarName As String) As Boolean

  'ASCII code 34 is double quotes
  SessionTableIsTouristName = ((Left(avatarName, 1) = Chr$(34)) And (Right(avatarName, 1) = Chr$(34)))
  
End Function
