Attribute VB_Name = "modAWSupport"
Option Explicit

Public mAWSDK                               As AwSdk

'***************************************************************************
'** Terrain texture rotation types                                        **
'**                                                                       **
'** Values for rotation are intrunsic but specified as follows:           **
'**                                                                       **
'**   AW_TERRAIN_ROTATE_NONE      &H00   00000000                         **
'**   AW_TERRAIN_ROTATE_LEFT      &H40   01000000                         **
'**   AW_TERRAIN_ROTATE_OPPOSITE  &H80   10000000                         **
'**   AW_TERRAIN_ROTATE_RIGHT     &HC0   11000000                         **
'**                                                                       **
'** These values are Logical OR'd with the texture value. If the overall  **
'** value of the texture is exactly 254, a hole is placed in the terrain. **
'***************************************************************************
Public Enum AW_TERRAIN_ROTATE_TYPE
  AW_TERRAIN_ROTATE_NONE = &H0
  AW_TERRAIN_ROTATE_LEFT = &H40
  AW_TERRAIN_ROTATE_OPPOSITE = &H80
  AW_TERRAIN_ROTATE_RIGHT = &HC0
End Enum

Public Enum AW_CELL_CALC_TYPE
  AW_CELL_CALC_VERTEX
  AW_CELL_CALC_CELL
End Enum

Public Type STRUCT_AW_TERRAIN_SET_HEIGHTS
  Heights(1 To 1000)        As Integer
End Type

'***************************************************************************
'** Global constants                                                      **
'***************************************************************************
Public Const AW_RC_OK                       As Long = 0
Public AW_TERRAIN_TEXTURE_MIN               As Long '= 0
Public AW_TERRAIN_TEXTURE_MAX               As Long ' = 62 changed to make option
Public Const AW_TERRAIN_TEXTURE_HOLE        As Long = 254

'***************************************************************************
'** Calculate the correct cell iterator value for a specific cell         **
'***************************************************************************
Public Function AWBuildCellIterator(ByVal CellX As Long, ByVal CellZ As Long) As Long
  
  AWBuildCellIterator = "&H" & Right("0000" & Hex(CellX), 4) & Right("0000" & Hex(CellZ), 4)

End Function

'***************************************************************************
'** Build the correct terrain texture value (assumes valid values)        **
'***************************************************************************
Public Function AWTerrainBuildTexture(ByVal textureHole As Boolean, ByVal textureIndex As Long, ByVal textureRotate As AW_TERRAIN_ROTATE_TYPE)

  Dim textureValue          As Long
  
  If textureHole Then
    textureValue = AW_TERRAIN_TEXTURE_HOLE
    
  Else
    textureValue = (textureIndex Or textureRotate)
    
  End If

  AWTerrainBuildTexture = textureValue
  
End Function


'***************************************************************************
'** Format a cell's location in AW representation                         **
'***************************************************************************
Public Function AWLocationNarrative(ByVal awCoordX As Long, ByVal awCoordZ As Long, Optional awAltitude As Long = 0, Optional awYaw As Long = 0) As String

  Dim awLocation            As String
  
  awLocation = Abs(awCoordZ / 1000) & IIf(awCoordZ < 0, "S", "N") & " " _
             & Abs(awCoordX / 1000) & IIf(awCoordX < 0, "E", "W")
             
  If Not ((awAltitude = 0) And (awYaw = 0)) Then
    awLocation = awLocation & " " & (awAltitude / 1000) & "a"
                        
    If Not (awAltitude = 0) Then
      awLocation = awLocation & " " & (awYaw / 10)
    
    End If
  End If
  
  AWLocationNarrative = awLocation
  
End Function


'***************************************************************************
'** Calculate the correct cell coord from an object position coord        **
'***************************************************************************
Public Function AWCellFromCoord(ByVal objectCoord As Long, ByVal cellCalcType As AW_CELL_CALC_TYPE) As Long

  Dim cellOffset            As Long
  Dim cellCoord             As Long
  
  cellOffset = objectCoord Mod 1000
  cellCoord = (objectCoord - cellOffset) / 1000
  
  If cellCalcType = AW_CELL_CALC_VERTEX Then
    If (Abs(cellOffset) > 500) Then
      cellCoord = cellCoord + Sgn(cellOffset)
  
    End If
  End If

  AWCellFromCoord = cellCoord
  
End Function

Public Function HasRight(ByVal Citnum As Long, ByVal WorldRight As Long)
    'This function checks to see if an avatar has a certain right
    'The rights are passed using the SDK enums - i.e. WorldRight = AW_WORLD_BUILD_RIGHT (46)
    
    Dim RightsList As String
    
    RightsList = mAWSDK.aw_string(WorldRight)
    
    'Check for occurrance of citizen number in string of the right
    If InStr(1, RightsList, Mid$(Str$(Citnum), 2), vbTextCompare) = 0 Then
        HasRight = False
    Else
        HasRight = True
    End If
    
    If InStr(1, RightsList, "*", vbTextCompare) <> 0 Then
        HasRight = True
    End If
    
End Function


