Attribute VB_Name = "modFormSupport"
Option Explicit

'System tray declarations begin:
'------------------------------
Const MAX_TOOLTIP As Integer = 64
Const NIF_ICON = &H2
Const NIF_MESSAGE = &H1
Const NIF_TIP = &H4
Const NIM_ADD = &H0
Const NIM_MODIFY = &H1
Const NIM_DELETE = &H2
Const WM_MOUSEMOVE = &H200
Const WM_LBUTTONDOWN = &H201
Const WM_LBUTTONUP = &H202
Const WM_LBUTTONDBLCLK = &H203
Const WM_RBUTTONDOWN = &H204
Const WM_RBUTTONUP = &H205
Const WM_RBUTTONDBLCLK = &H206

Private Type NOTIFYICONDATA
    cbSize           As Long
    hwnd             As Long
    uID              As Long
    uFlags           As Long
    uCallbackMessage As Long
    hIcon            As Long
    szTip            As String * MAX_TOOLTIP
End Type
Private nfIconData As NOTIFYICONDATA

Private Declare Function Shell_NotifyIcon Lib "shell32.dll" Alias "Shell_NotifyIconA" (ByVal dwMessage As Long, lpData As NOTIFYICONDATA) As Long

'---------------------------------
'End system tray declarations

Public Const TERRAIN_TRACK_COUNTERS         As Boolean = True

Public Enum FORM_STATE_TYPE
  FORM_STATE_READY
  FORM_STATE_LOGGING_IN
  FORM_STATE_ACTIVE
  FORM_STATE_TERMINATED
  FORM_STATE_DISCONNECTED
End Enum

Public Enum FORM_CONNECTION_STATE
  FORM_CONNECTION_ONLINE
  FORM_CONNECTION_OFFLINE
  FORM_CONNECTION_DISCONNECTED
End Enum

Public Const FORM_LOG_ERROR     As String = "Error"
Public Const FORM_LOG_ACTION    As String = "Action"
Public Const FORM_LOG_COMMAND   As String = "Command"
Public Const FORM_LOG_SAFEGUARD As String = "Safeguard"
Public Const FORM_LOG_MESSAGE   As String = "Message"

Public FORM_CURRENT_STATE                   As FORM_STATE_TYPE

Public Const TERRAIN_AWWAIT_INTERVAL        As Long = 50
Public Const TERRAIN_QUEUE_INTERVAL         As Long = 10
Public Const TERRAIN_AUTO_INTERVAL          As Long = 10000

Public Const LB_SETTABSTOPS                 As Long = &H192

Public Declare Function SendMessage Lib "user32" Alias "SendMessageA" (ByVal hwnd As Long, ByVal wMsg As Long, ByVal wParam As Long, lParam As Any) As Long

Public Sub FormSaveLogOptions()

  With frmTerrainManager

    WriteToINI .chkFileLog, "Logging", "Enabled", IniPath()
    WriteToINI .txtFileLog, "Logging", "File", IniPath()
    
  End With

End Sub

Public Sub FormGetLogOptions()

  With frmTerrainManager

    .chkFileLog = GetFromINI("Logging", "Enabled", IniPath(), "0")
    .txtFileLog = GetFromINI("Logging", "File", IniPath(), LogPath())
    
  End With

End Sub

Public Sub FormListInitialize()

  With frmTerrainManager.ActivityLog
    .ColumnHeaders.Add , , "Date/Time"
    .ColumnHeaders.Add , , "Type"
    .ColumnHeaders.Add , , "Privilege"
    .ColumnHeaders.Add , , "Name"
    .ColumnHeaders.Add , , "Message"
    .ColumnHeaders(1).Width = .Width / 3.7
    .ColumnHeaders(2).Width = .Width / 7
    .ColumnHeaders(3).Width = .Width / 7
    .ColumnHeaders(4).Width = .Width / 7
    .ColumnHeaders(5).Width = .Width / 3.5
  End With

End Sub

Public Sub FormDeleteFromSystemTray()

    Call Shell_NotifyIcon(NIM_DELETE, nfIconData)

End Sub
Public Sub FormDisableAutoLogin()

  'Disable login checkbox, save changes to INI, and restore form to visible
  With frmTerrainManager
    
    'Nofity user that auto-login is disabled
    If .chkAutoLogin = 1 Then
      FormWriteToLog FORM_LOG_MESSAGE, , , "Auto-login has been disabled."
      
    End If
  
    WriteToINI 0, "World", "AutoLogin", IniPath()
    .chkAutoLogin = 0
    .mnuRestore_Click
  
  End With
  
End Sub

Public Sub FormMouseMove(ByVal Button As Integer, ByVal Shift As Integer, ByVal X As Single, ByVal Y As Single)

Dim lMsg As Single
'
' Determine the event that happened to the System Tray icon.
' Left clicking the icon displays a message box.
' Right clicking the icon creates an instance of an object from an
' ActiveX Code component then invokes a method to display a message.
'

'No need to check anything here if the form isn't minimized
If frmTerrainManager.WindowState <> 1 Then Exit Sub

lMsg = X / Screen.TwipsPerPixelX
Select Case lMsg
    Case WM_LBUTTONUP
        frmTerrainManager.mnuRestore_Click 'restore form
    Case WM_RBUTTONUP
        'popup menu
        frmTerrainManager.PopupMenu frmTerrainManager.mnuPopup, , , , frmTerrainManager.mnuRestore
    'Case WM_MOUSEMOVE
    'Case WM_LBUTTONDOWN
    'Case WM_LBUTTONDBLCLK
    'Case WM_RBUTTONDOWN
    'Case WM_RBUTTONDBLCLK
    'Case Else
End Select

End Sub


Public Sub FormAddToSystemTray()

    ' Add this application's icon to the system tray.
    '
    ' Parm 1 = Handle of the window to receive notification messages
    '          associated with an icon in the taskbar status area.
    ' Parm 2 = Icon to display.
    ' Parm 3 = Handle of icon to display.
    ' Parm 4 = Tooltip displayed when cursor moves over system tray icon.
    '
    With nfIconData
        .hwnd = frmTerrainManager.hwnd
        .uID = frmTerrainManager.Icon
        .uFlags = NIF_ICON Or NIF_MESSAGE Or NIF_TIP
        .uCallbackMessage = WM_MOUSEMOVE
        .hIcon = frmTerrainManager.Icon.Handle
        .cbSize = Len(nfIconData)
    End With
    Call Shell_NotifyIcon(NIM_ADD, nfIconData)

End Sub
    
Public Sub LogoutWithError(ErrorMessage As String)

  FormWriteToLog FORM_LOG_ERROR, , , ErrorMessage
  FormSetState FORM_STATE_TERMINATED
  FormSetConnectState FORM_CONNECTION_OFFLINE
  mAWSDK.aw_destroy
  mAWSDK.aw_term
  FormDisableAutoLogin
  Exit Sub
  
End Sub

Public Sub FormLogoutWithErrorAndRecover(ErrorMessage As String)

  FormWriteToLog FORM_LOG_ERROR, , , ErrorMessage
  FormSetState FORM_STATE_READY
  FormSetConnectState FORM_CONNECTION_OFFLINE
  mAWSDK.aw_destroy
  FormDisableAutoLogin
  Exit Sub
  
End Sub

Public Function IniPath()
  'This function fixes an undesirable behavior in VB6 where the
  'app.path property is inconsistent
  If Right$(App.Path, 1) = "\" Then IniPath = App.Path + "demeter.ini"
  If Right$(App.Path, 1) <> "\" Then IniPath = App.Path + "\demeter.ini"
End Function

Public Function LogPath()
  'This function fixes an undesirable behavior in VB6 where the
  'app.path property is inconsistent
  If Right$(App.Path, 1) = "\" Then LogPath = App.Path + "log.txt"
  If Right$(App.Path, 1) <> "\" Then LogPath = App.Path + "\log.txt"
End Function

Public Sub SaveLoginOptions()

  With frmTerrainManager
    WriteToINI .txtPassword, "Login", "PrivilegePassword", IniPath()
    WriteToINI .txtTargetWorld, "Login", "World", IniPath()
    WriteToINI .txtUserId, "Login", "CitizenNumber", IniPath()
    WriteToINI .txtUniHost, "Login", "UniHost", IniPath()
    WriteToINI .txtUniPort, "Login", "UniPort", IniPath()
    WriteToINI .chkLoginQuery, "Login", "LoginQuery", IniPath()
  End With

End Sub



Public Sub FormSetState(ByVal formState As FORM_STATE_TYPE, Optional altStateText As String = vbNullString)

  Dim loginEnabled          As Boolean
  Dim logEnabled            As Boolean
  Dim stateText             As String
  
  With frmTerrainManager
    Select Case formState
      Case FORM_STATE_READY
        stateText = "Ready"
        loginEnabled = True
        logEnabled = False
        
      Case FORM_STATE_LOGGING_IN
        stateText = "Logging In"
        loginEnabled = False
        logEnabled = False
      
      Case FORM_STATE_ACTIVE
        stateText = "Active"
        loginEnabled = False
        logEnabled = True
        
      Case FORM_STATE_TERMINATED
        stateText = "Terminated: Restart Required"
        loginEnabled = False
        logEnabled = False
        
      Case FORM_STATE_DISCONNECTED
        stateText = "Disconnected; Reconnecting..."
        loginEnabled = False
        logEnabled = False
      
    End Select
    
    FORM_CURRENT_STATE = formState
    
    .lblTitleTargetWorld.Enabled = loginEnabled
    .lblTitleCitNum.Enabled = loginEnabled
    .lblTitlePassword.Enabled = loginEnabled
    .lblUniHost.Enabled = loginEnabled
    .lblUniPort.Enabled = loginEnabled
    .chkLoginQuery.Enabled = loginEnabled
    
    .txtTargetWorld.Enabled = loginEnabled
    .txtUserId.Enabled = loginEnabled
    .txtPassword.Enabled = loginEnabled
    .txtUniHost.Enabled = loginEnabled
    .txtUniPort.Enabled = loginEnabled
    .fLogin.Enabled = loginEnabled
  
    .pbLogin.Enabled = loginEnabled
    .pbLogin.Default = loginEnabled

    '.ActivityLog.Enabled = logEnabled
    '.fLog.Enabled = logEnabled
    
    .lblState = IIf(altStateText = vbNullString, stateText, altStateText)
    .lblState.Refresh
    
    If loginEnabled And .Visible Then .txtPassword.SetFocus
    
    'update tray icon tool tip text
    FormUpdateIconToolTip

  End With
  
End Sub

Public Sub FormUpdateIconToolTip()
  
  Dim stateText As String

  'Check to see if tray icon is visible and update it to new text
  With frmTerrainManager
  
    'Nothing needs to be done here if the form is visible
    'The only time something is done is when the tray icon is visible and
    'The form is invisible
    If Not .Visible Then
    
      Select Case FORM_CURRENT_STATE
        Case FORM_STATE_READY
          stateText = "Ready"
        
        Case FORM_STATE_LOGGING_IN
          stateText = "Logging In"
          
        Case FORM_STATE_ACTIVE
          stateText = "Active"
          
        Case FORM_STATE_TERMINATED
          stateText = "Terminated"
          
        Case FORM_STATE_DISCONNECTED
          stateText = "Disconnected"
     
       End Select
    
       nfIconData.szTip = "Demeter - " + stateText + " - " + _
       "Markers:" + Left(.lblMarkerCount, Len(.lblMarkerCount) - 1) + _
       ", Avatars:" + Left(.lblAvatarCount, Len(.lblAvatarCount) - 1) + _
       ", Sessions:" + Left(.lblTerrainCount, Len(.lblTerrainCount) - 1) _
       + vbNullChar
       'All of this is necessary because spaces are inserted on the form
       'after the numbers in the labels
        
       Call Shell_NotifyIcon(NIM_MODIFY, nfIconData)
  
    End If
  
  End With
  
End Sub

Public Sub FormWriteToLog(ByVal logType As String, Optional ByVal privNum As Long = -1, Optional ByVal avName As String = "", Optional ByVal messageText As String = "")

  On Error GoTo ErrorHandler

  With frmTerrainManager.ActivityLog.ListItems
    If (.Count > 1000) Then .Remove .Count
    .Add 1, , CDate(Now)
  End With
  
  With frmTerrainManager.ActivityLog.ListItems(1)
    .SubItems(1) = logType
    
    If privNum = -1 Then 'not specified, this is an error
      .SubItems(2) = vbNullString
    Else
      .SubItems(2) = privNum
      
    End If
    
    .SubItems(3) = avName
    .SubItems(4) = messageText
  End With

  'Prevent infinite recursion
  If Left(messageText, 25) = "Unable to write to file -" Then
    Exit Sub

  End If
  
  With frmTerrainManager
    If .chkFileLog = 1 Then
      Open .txtFileLog For Append As #1
        Print #1, CStr(Now) + Space(25 - Len(Now)) + logType + Space(12 - Len(logType)) + IIf(privNum = -1, vbNullString, its(privNum)) + Space(8 - Len(IIf(privNum = -1, vbNullString, its(privNum)))) + avName + Space(20 - Len(avName)) + messageText
      
      Close #1
      
    End If
      
  End With
  
  Exit Sub
  
ErrorHandler:
  FormWriteToLog FORM_LOG_ERROR, , , "Unable to write to file - use the 'Browse' button to select a log file"
  Close #1

End Sub

Public Sub FormSetMarkerCount(ByVal markerCount As Long)

  If TERRAIN_TRACK_COUNTERS Then
    With frmTerrainManager.lblMarkerCount
      .Caption = markerCount & " "
      .Refresh
    
    End With
  End If
  
End Sub

Public Function FormCalcCellsInWorld()

  'Returns the number of cells in the world
   FormCalcCellsInWorld = (TERRAIN_WORLD_COORD_SIZE * 2 + 1) ^ 2

End Function


Public Sub FormSetCellCount(ByVal cellCount As Long)

  If TERRAIN_TRACK_COUNTERS Then
    With frmTerrainManager.lblCellCount
          .Caption = cellCount & " "
          .Refresh

    End With
  
    FormSetPercentQueried
    
  End If
  
End Sub

Public Function FormGetPercentQueried()

  'Gets the percentage queried for the world in string format
  FormGetPercentQueried = Format(TerrainGetCellCount() * 100 / _
                          FormCalcCellsInWorld(), "##0.000") & "% "

End Function

Public Sub FormSetPercentQueried()

  With frmTerrainManager.lblPercentQueried
    If TERRAIN_TRACK_COUNTERS Then
        'division calculates percentage of cells in world
        'that have been queried
        .Caption = FormGetPercentQueried()
        If .Caption = "100.000% " Then .ForeColor = RGB(255, 0, 0)
        .Refresh
    
     Else
       'avoid divide by zero error if not tracking
       .Caption = " "
      
     End If
     
  End With

End Sub

Public Sub FormSetAvatarCount(ByVal avatarCount As Long)

  With frmTerrainManager.lblAvatarCount
    .Caption = avatarCount & " "
    .Refresh
    
  End With
  
End Sub

Public Sub FormSetTerrainSessionCount(ByVal terrainSessionCount As Long)

  With frmTerrainManager.lblTerrainCount
    .Caption = terrainSessionCount & " "
    .Refresh
    
  End With

End Sub

Public Sub FormSetCellsInWorld()

  With frmTerrainManager.lblCellsInWorld
    'square of width of world in cells
    .Caption = FormCalcCellsInWorld() & " "
    .Refresh
    
  End With
  
End Sub


Public Sub FormSetWorldName(ByVal worldName As String)

  With frmTerrainManager.lblWorld
    .Caption = worldName & " "
    .Refresh
    
  End With
  
End Sub

Public Sub FormSetConnectState(ByVal connectState As FORM_CONNECTION_STATE)

  With frmTerrainManager.lblConnectState
    
    Select Case connectState
    Case FORM_CONNECTION_ONLINE
        .Caption = "Online "
    Case FORM_CONNECTION_OFFLINE
        .Caption = "Offline "
    Case FORM_CONNECTION_DISCONNECTED
        .Caption = "Disconnected "
    End Select
    
    .Refresh
    
  End With
  
End Sub


