Attribute VB_Name = "modFormSupport"
Option Explicit

'System tray declarations begin:
'------------------------------
Const MAX_TOOLTIP As Integer = 64
Const NIF_ICON = &H2
Const NIF_MESSAGE = &H1
Const NIF_TIP = &H4
Const NIM_ADD = &H0
Const NIM_MODIFY = &H1
Const NIM_DELETE = &H2
Const WM_MOUSEMOVE = &H200
Const WM_LBUTTONDOWN = &H201
Const WM_LBUTTONUP = &H202
Const WM_LBUTTONDBLCLK = &H203
Const WM_RBUTTONDOWN = &H204
Const WM_RBUTTONUP = &H205
Const WM_RBUTTONDBLCLK = &H206

Private Type NOTIFYICONDATA
    cbSize           As Long
    hwnd             As Long
    uID              As Long
    uFlags           As Long
    uCallbackMessage As Long
    hIcon            As Long
    szTip            As String * MAX_TOOLTIP
End Type
Private nfIconData As NOTIFYICONDATA

Private Declare Function Shell_NotifyIcon Lib "shell32.dll" Alias "Shell_NotifyIconA" (ByVal dwMessage As Long, lpData As NOTIFYICONDATA) As Long

'---------------------------------
'End system tray declarations

Public Const TERRAIN_TRACK_COUNTERS         As Boolean = True

Public Enum FORM_STATE_TYPE
  FORM_STATE_READY
  FORM_STATE_LOGGING_IN
  FORM_STATE_ACTIVE
  FORM_STATE_TERMINATED
  FORM_STATE_DISCONNECTED
End Enum

Public Enum FORM_CONNECTION_STATE
  FORM_CONNECTION_ONLINE
  FORM_CONNECTION_OFFLINE
  FORM_CONNECTION_DISCONNECTED
End Enum

Public FORM_CURRENT_STATE                   As FORM_STATE_TYPE

Public Const TERRAIN_AWWAIT_INTERVAL        As Long = 50
Public Const TERRAIN_QUEUE_INTERVAL         As Long = 1
Public Const TERRAIN_AUTO_INTERVAL          As Long = 10000

Public Const LB_SETTABSTOPS                 As Long = &H192

Public Declare Function SendMessage Lib "user32" Alias "SendMessageA" (ByVal hwnd As Long, ByVal wMsg As Long, ByVal wParam As Long, lParam As Any) As Long

Public Sub FormDeleteFromSystemTray()

    Call Shell_NotifyIcon(NIM_DELETE, nfIconData)

End Sub
Public Sub FormMouseMove(ByVal Button As Integer, ByVal Shift As Integer, ByVal X As Single, ByVal Y As Single)

Dim lMsg As Single
'
' Determine the event that happened to the System Tray icon.
' Left clicking the icon displays a message box.
' Right clicking the icon creates an instance of an object from an
' ActiveX Code component then invokes a method to display a message.
'
lMsg = X / Screen.TwipsPerPixelX
Select Case lMsg
    Case WM_LBUTTONUP
        frmTerrainManager.mnuRestore_Click 'restore form
    Case WM_RBUTTONUP
        'popup menu
        frmTerrainManager.PopupMenu frmTerrainManager.mnuPopup, , , , frmTerrainManager.mnuRestore
    'Case WM_MOUSEMOVE
    'Case WM_LBUTTONDOWN
    'Case WM_LBUTTONDBLCLK
    'Case WM_RBUTTONDOWN
    'Case WM_RBUTTONDBLCLK
    'Case Else
End Select

End Sub


Public Sub FormAddToSystemTray()

    ' Add this application's icon to the system tray.
    '
    ' Parm 1 = Handle of the window to receive notification messages
    '          associated with an icon in the taskbar status area.
    ' Parm 2 = Icon to display.
    ' Parm 3 = Handle of icon to display.
    ' Parm 4 = Tooltip displayed when cursor moves over system tray icon.
    '
    With nfIconData
        .hwnd = frmTerrainManager.hwnd
        .uID = frmTerrainManager.Icon
        .uFlags = NIF_ICON Or NIF_MESSAGE Or NIF_TIP
        .uCallbackMessage = WM_MOUSEMOVE
        .hIcon = frmTerrainManager.Icon.Handle
        .cbSize = Len(nfIconData)
    End With
    Call Shell_NotifyIcon(NIM_ADD, nfIconData)

End Sub
    
Public Sub LogoutWithError(ErrorMessage As String)

  FormWriteToLog "[ERROR] " + ErrorMessage
  FormSetState FORM_STATE_TERMINATED
  FormSetConnectState FORM_CONNECTION_OFFLINE
  mAWSDK.aw_destroy
  mAWSDK.aw_term
  Exit Sub
  
End Sub

Public Function IniPath()
  'This function fixes an undesirable behavior in VB6 where the
  'app.path property is inconsistent
  If Right$(App.Path, 1) = "\" Then IniPath = App.Path + "demeter.ini"
  If Right$(App.Path, 1) <> "\" Then IniPath = App.Path + "\demeter.ini"
End Function

Public Sub SaveLoginOptions()
  With frmTerrainManager
    .ini.WriteToINI .txtPassword, "Login", "PrivilegePassword", IniPath()
    .ini.WriteToINI .txtTargetWorld, "Login", "World", IniPath()
    .ini.WriteToINI .txtUserId, "Login", "CitizenNumber", IniPath()
    .ini.WriteToINI .txtUniHost, "Login", "UniHost", IniPath()
    .ini.WriteToINI .txtUniPort, "Login", "UniPort", IniPath()
  End With
End Sub



Public Sub FormSetState(ByVal formState As FORM_STATE_TYPE, Optional altStateText As String = vbNullString)

  Dim loginEnabled          As Boolean
  Dim logEnabled            As Boolean
  Dim stateText             As String
  
  With frmTerrainManager
    Select Case formState
      Case FORM_STATE_READY
        stateText = "Ready"
        loginEnabled = True
        logEnabled = False
        
      Case FORM_STATE_LOGGING_IN
        stateText = "Logging In"
        loginEnabled = False
        logEnabled = False
      
      Case FORM_STATE_ACTIVE
        stateText = "Active"
        loginEnabled = False
        logEnabled = True
        
      Case FORM_STATE_TERMINATED
        stateText = "Terminated: Restart Required"
        loginEnabled = False
        logEnabled = False
        
      Case FORM_STATE_DISCONNECTED
        stateText = "Disconnected; Reconnecting..."
        loginEnabled = False
        logEnabled = False
      
    End Select
    
    FORM_CURRENT_STATE = formState
    
    .lblTitleTargetWorld.Enabled = loginEnabled
    .lblTitleCitNum.Enabled = loginEnabled
    .lblTitlePassword.Enabled = loginEnabled
    .lblUniHost.Enabled = loginEnabled
    .lblUniPort.Enabled = loginEnabled
    .txtTargetWorld.Enabled = loginEnabled
    .txtUserId.Enabled = loginEnabled
    .txtPassword.Enabled = loginEnabled
    .txtUniHost.Enabled = loginEnabled
    .txtUniPort.Enabled = loginEnabled
    .fLogin.Enabled = loginEnabled
  
    .pbLogin.Enabled = loginEnabled
    .pbLogin.Default = loginEnabled
    
    .pbClear.Enabled = logEnabled
    
    .pbStatistics.Enabled = logEnabled
    .pbCacheDump.Enabled = logEnabled
    .lbLog.Enabled = logEnabled
    .fLog.Enabled = logEnabled
    
    .lblState = IIf(altStateText = vbNullString, stateText, altStateText)
    .lblState.Refresh
    
    If loginEnabled And .Visible Then .txtPassword.SetFocus
    
    'update tray icon tool tip text
    FormUpdateIconToolTip

  End With
  
End Sub

Public Sub FormUpdateIconToolTip()
  
  Dim stateText As String

  'Check to see if tray icon is visible and update it to new text
  With frmTerrainManager
  
    'Nothing needs to be done here if the form is visible
    'The only time something is done is when the tray icon is visible and
    'The form is invisible
    If Not .Visible Then
    
      Select Case FORM_CURRENT_STATE
        Case FORM_STATE_READY
          stateText = "Ready"
        
        Case FORM_STATE_LOGGING_IN
          stateText = "Logging In"
          
        Case FORM_STATE_ACTIVE
          stateText = "Active"
          
        Case FORM_STATE_TERMINATED
          stateText = "Terminated"
          
        Case FORM_STATE_DISCONNECTED
          stateText = "Disconnected"
     
       End Select
    
       nfIconData.szTip = "Demeter - " + stateText + " - " + _
       "Markers:" + Left(.lblMarkerCount, Len(.lblMarkerCount) - 1) + _
       ", Avatars:" + Left(.lblAvatarCount, Len(.lblAvatarCount) - 1) + _
       ", Sessions:" + Left(.lblTerrainCount, Len(.lblTerrainCount) - 1) _
       + vbNullChar
       'All of this is necessary because spaces are inserted on the form
       'after the numbers in the labels
        
       Call Shell_NotifyIcon(NIM_MODIFY, nfIconData)
  
    End If
  
  End With
  
End Sub

Public Sub FormWriteToLog(ByVal logText As String)

  With frmTerrainManager.lbLog
    If (.ListCount > 1000) Then .RemoveItem (.ListCount - 1)
    .AddItem Format(Now(), "DD/MM/YYYY HH:MM:SS") & vbTab & logText, 0
    
  End With
  
End Sub

Public Sub FormSetMarkerCount(ByVal markerCount As Long)

  If TERRAIN_TRACK_COUNTERS Then
    With frmTerrainManager.lblMarkerCount
      .Caption = markerCount & " "
      .Refresh
    
    End With
  End If
  
End Sub

Public Sub FormSetCellCount(ByVal cellCount As Long)

  If TERRAIN_TRACK_COUNTERS Then
    With frmTerrainManager.lblCellCount
      .Caption = cellCount & " "
      .Refresh
    
    End With
  End If
  
End Sub

Public Sub FormSetAvatarCount(ByVal avatarCount As Long)

  With frmTerrainManager.lblAvatarCount
    .Caption = avatarCount & " "
    .Refresh
    
  End With
  
End Sub

Public Sub FormSetTerrainSessionCount(ByVal terrainSessionCount As Long)

  With frmTerrainManager.lblTerrainCount
    .Caption = terrainSessionCount & " "
    .Refresh
    
  End With

End Sub

Public Sub FormSetWorldName(ByVal worldName As String)

  With frmTerrainManager.lblWorld
    .Caption = worldName & " "
    .Refresh
    
  End With
  
End Sub

Public Sub FormSetConnectState(ByVal connectState As FORM_CONNECTION_STATE)

  With frmTerrainManager.lblConnectState
    
    Select Case connectState
    Case FORM_CONNECTION_ONLINE
        .Caption = "Online "
    Case FORM_CONNECTION_OFFLINE
        .Caption = "Offline "
    Case FORM_CONNECTION_DISCONNECTED
        .Caption = "Disconnected "
    End Select
    
    .Refresh
    
  End With
  
End Sub


