Attribute VB_Name = "modFormSupport"
Option Explicit

Public Const TERRAIN_TRACK_COUNTERS         As Boolean = True

Public Enum FORM_STATE_TYPE
  FORM_STATE_READY
  FORM_STATE_LOGGING_IN
  FORM_STATE_ACTIVE
  FORM_STATE_TERMINATED
  FORM_STATE_DISCONNECTED
End Enum

Public Enum FORM_CONNECTION_STATE
  FORM_CONNECTION_ONLINE
  FORM_CONNECTION_OFFLINE
  FORM_CONNECTION_DISCONNECTED
End Enum

Public Const TERRAIN_AWWAIT_INTERVAL        As Long = 50
Public Const TERRAIN_QUEUE_INTERVAL         As Long = 1
Public Const TERRAIN_AUTO_INTERVAL          As Long = 10000

Public Const LB_SETTABSTOPS                 As Long = &H192

Public Declare Function SendMessage Lib "User32" Alias "SendMessageA" (ByVal hWnd As Long, ByVal wMsg As Long, ByVal wParam As Long, lParam As Any) As Long

Public Sub LogoutWithError(ErrorMessage As String)
  FormWriteToLog "[ERROR] " + ErrorMessage
  FormSetState FORM_STATE_TERMINATED
  FormSetConnectState FORM_CONNECTION_OFFLINE
  mAWSDK.aw_destroy
  mAWSDK.aw_term
  Exit Sub
End Sub

Public Function IniPath()
  If Right$(App.Path, 1) = "\" Then IniPath = App.Path + "demeter.ini"
  If Right$(App.Path, 1) <> "\" Then IniPath = App.Path + "\demeter.ini"
End Function

Public Sub SaveLoginOptions()
  With frmTerrainManager
    .ini.WriteToINI .txtPassword, "Login", "PrivilegePassword", IniPath()
    .ini.WriteToINI .txtTargetWorld, "Login", "World", IniPath()
    .ini.WriteToINI .txtUserId, "Login", "CitizenNumber", IniPath()
    .ini.WriteToINI .txtUniHost, "Login", "UniHost", IniPath()
    .ini.WriteToINI .txtUniPort, "Login", "UniPort", IniPath()
  End With
End Sub

Public Function FormGetState()
  'Returns current state of form based on text assigned to stateText label
  
  With frmTerrainManager
    Select Case .lblState
      Case "Ready"
        FormGetState = FORM_STATE_READY
      Case "Logging In"
        FormGetState = FORM_STATE_LOGGING_IN
      Case "Active"
        FormGetState = FORM_STATE_ACTIVE
      Case "Terminated: Restart Required"
        FormGetState = FORM_STATE_TERMINATED
      Case "Disconnected; Reconnecting..."
        FormGetState = FORM_STATE_DISCONNECTED
    End Select
  End With
  
End Function

Public Sub FormSetState(ByVal formState As FORM_STATE_TYPE, Optional altStateText As String = vbNullString)

  Dim loginEnabled          As Boolean
  Dim logEnabled            As Boolean
  Dim stateText             As String
  
  With frmTerrainManager
    Select Case formState
      Case FORM_STATE_READY
        stateText = "Ready"
        loginEnabled = True
        logEnabled = False
        
      Case FORM_STATE_LOGGING_IN
        stateText = "Logging In"
        loginEnabled = False
        logEnabled = False
      
      Case FORM_STATE_ACTIVE
        stateText = "Active"
        loginEnabled = False
        logEnabled = True
        
      Case FORM_STATE_TERMINATED
        stateText = "Terminated: Restart Required"
        loginEnabled = False
        logEnabled = False
        
      Case FORM_STATE_DISCONNECTED
        stateText = "Disconnected; Reconnecting..."
        loginEnabled = False
        logEnabled = False
      
    End Select
    
    .lblTitleTargetWorld.Enabled = loginEnabled
    .lblTitleCitNum.Enabled = loginEnabled
    .lblTitlePassword.Enabled = loginEnabled
    .lblUniHost.Enabled = loginEnabled
    .lblUniPort.Enabled = loginEnabled
    .txtTargetWorld.Enabled = loginEnabled
    .txtUserId.Enabled = loginEnabled
    .txtPassword.Enabled = loginEnabled
    .txtUniHost.Enabled = loginEnabled
    .txtUniPort.Enabled = loginEnabled
    .fLogin.Enabled = loginEnabled
  
    .pbLogin.Enabled = loginEnabled
    .pbLogin.Default = loginEnabled
    
    .pbClear.Enabled = logEnabled
    
    .pbStatistics.Enabled = logEnabled
    .pbCacheDump.Enabled = logEnabled
    .lbLog.Enabled = logEnabled
    .fLog.Enabled = logEnabled
    
    .lblState = IIf(altStateText = vbNullString, stateText, altStateText)
    .lblState.Refresh
    
    If loginEnabled Then .txtPassword.SetFocus

  End With
  
End Sub

Public Sub FormWriteToLog(ByVal logText As String)

  With frmTerrainManager.lbLog
    If (.ListCount > 1000) Then .RemoveItem (.ListCount - 1)
    .AddItem Format(Now(), "DD/MM/YYYY HH:MM:SS") & vbTab & logText, 0
    
  End With
  
End Sub

Public Sub FormSetMarkerCount(ByVal markerCount As Long)

  If TERRAIN_TRACK_COUNTERS Then
    With frmTerrainManager.lblMarkerCount
      .Caption = markerCount & " "
      .Refresh
    
    End With
  End If
  
End Sub

Public Sub FormSetCellCount(ByVal cellCount As Long)

  If TERRAIN_TRACK_COUNTERS Then
    With frmTerrainManager.lblCellCount
      .Caption = cellCount & " "
      .Refresh
    
    End With
  End If
  
End Sub

Public Sub FormSetAvatarCount(ByVal avatarCount As Long)

  With frmTerrainManager.lblAvatarCount
    .Caption = avatarCount & " "
    .Refresh
    
  End With
  
End Sub

Public Sub FormSetTerrainSessionCount(ByVal terrainSessionCount As Long)

  With frmTerrainManager.lblTerrainCount
    .Caption = terrainSessionCount & " "
    .Refresh
    
  End With

End Sub

Public Sub FormSetWorldName(ByVal worldName As String)

  With frmTerrainManager.lblWorld
    .Caption = worldName & " "
    .Refresh
    
  End With
  
End Sub

Public Sub FormSetConnectState(ByVal connectState As FORM_CONNECTION_STATE)

  With frmTerrainManager.lblConnectState
    
    Select Case connectState
    Case FORM_CONNECTION_ONLINE
        .Caption = "Online "
    Case FORM_CONNECTION_OFFLINE
        .Caption = "Offline "
    Case FORM_CONNECTION_DISCONNECTED
        .Caption = "Disconnected "
    End Select
    
    .Refresh
    
  End With
  
End Sub


